/*  Phase Calibrator for emonTx & emonTx Shield
    Based on a single phase energy diverter by Martin Roberts 2/12/12, 
    this implements a phase-locked loop to synchronise to the mains supply.
    
*/
    
//--------------------------------------------------------------------------------------------------
// These are the three settings that must be checked:


#define EMONTX_V34                               // Sets the I/O pin allocation. 
                                                 // Use EMONTX_V22 or EMONTX25 or EMONTX_V32
                                                 // or EMONTX_V34 or EMONTX_SHIELD as appropriate.
                                                 // Note that current pins can be chosen at run time.
#define SUPPLY_FREQUENCY 50                      // Change to 60 if appropriate

#define SUPPLY_VOLTS 3.3                         // Use 5.0 for Arduino / emonTx Shield

//--------------------------------------------------------------------------------------------------
// #define DEBUGGING                             // enable this line to include debugging print statements



//--------------------------------------------------------------------------------------------------
#define VERSION "PLL Phase Calibrator V1.0"                      


#define LEDISLOCK         // comment this out for LED pulsed during transmission
                          //  otherwise LED shows loop is locked, and occults to show data ready, but that is not easily visible

//--------------------------------------------------------------------------------------------------
// other system constants that might need to be changed to suit a different processor or ADC

#define NUMSAMPLES 36     
#define CTPhase NUMSAMPLES/4

#define ADC_BITS 10       // ADC Resolution
#define ADC_RATE 64       // Time between successive ADC conversions in microseconds

#define PLLTIMERRANGE 100 // PLL timer range limit ~ +/-0.5Hz
#define PLLLOCKRANGE 40   // allowable ADC range to enter locked state
#define PLLUNLOCKRANGE 80 // allowable ADC range to remain locked
#define PLLLOCKCOUNT 100  // number of cycles to determine if PLL is locked

//--------------------------------------------------------------------------------------------------
//
//   Users should not need to change anything below here
//
//--------------------------------------------------------------------------------------------------

#include <SPI.h>


//--------------------------------------------------------------------------------------------------
// I/O pins for debugging
// Define these only for testing. On the emonTx V3.x, it may be necessary to comment out and
//  remove external connections to the interrupt pin, the One-wire or DS18B20 power pins.
// #define SYNCPIN 5  // this output will be a 50Hz square wave locked to the 50Hz input
// #define SAMPPIN 4   // this output goes high each time an ADC conversion starts or completes
//#define TXPIN 3     // this output goes high each time a data output takes place
//--------------------------------------------------------------------------------------------------

// Arduino I/O pin usage
#if defined(EMONTX_V22)
// EmonTx v2.2 Pin references
#define VOLTSPIN 2
#define CT1PIN 3
#define LEDPIN 9
#define VCAL 227.6
#define ICAL 111.1

#elif defined(EMONTX_V25)
// EmonTx v2.5 Pin references
#define VOLTSPIN 0
#define CT1PIN 1
#define LEDPIN 9
#define VCAL 269.0
#define ICAL 90.91

#elif defined(EMONTX_V32)
// EmonTx v3.2 Pin references
#define VOLTSPIN 0
#define CT1PIN 1
#define LEDPIN 6
#define VCAL 269.0
#define ICAL 90.91

#elif defined EMONTX_SHIELD

// EmonTx Shield Pin references
#define VOLTSPIN 0
#define CT1PIN 1
#define LEDPIN 9
#define VCAL 269.0
#define ICAL 60.6

#else
// EmonTx v3.4 Pin references
#define VOLTSPIN 0
#define CT1PIN 1
#define LEDPIN 6
#define VCAL 269.0
#define ICAL 90.91
#define DIP_SWITCH1 8   // [Not used]
#define DIP_SWITCH2 9   // [Not used]

#endif

//-------------------------------------------------------------------------------------------------
//  Starting Values
//-------------------------------------------------------------------------------------------------
  
byte ctPin  = CT1PIN;
double vCal = VCAL;
double iCal = ICAL; 

double iLead = 0.0;
double iError = 0.072;   //  system error in degrees (~4 µs)

//--------------------------------------------------------------------------------------------------

//--------------------------------------------------------------------------------------------------
// constants calculated at compile time
#define BUFFERSIZE (CTPhase + 2)                  // Store a little more than 90 degrees of voltage samples

#define ADC_COUNTS (1 << ADC_BITS)               // ADC Resolution in steps
#define SAMPLERATE (360.0 / NUMSAMPLES)          // Sample Rate in degrees


#define TIMERTOP (((1000000/SUPPLY_FREQUENCY/NUMSAMPLES)*16)-1) // terminal count for PLL timer
#define PLLTIMERMAX (TIMERTOP+PLLTIMERRANGE)
#define PLLTIMERMIN (TIMERTOP-PLLTIMERRANGE)
  
//--------------------------------------------------------------------------------------------------


double looptime = 2000;                          // time of outer loop in milliseconds, also time between data transmissions


// Intermediate constants
double v_ratio, i_ratio;
double phaseshift;

bool firstCycle = true; 

// Accumulated values over 1 cycle - shared between ISR & main program
volatile unsigned long sumVsq, sumIsq; 
volatile long sumVavg, sumIavg;
volatile long sumPowerA, sumPowerB; 
volatile unsigned int sumSamples;

// Accumulated values over the reporting period
uint64_t sumPeriodVsq, sumPeriodIsq;
int64_t sumPeriodVavg, sumPeriodIavg;
int64_t sumPeriodPowerA, sumPeriodPowerB; 
unsigned long sumPeriodSamples;

double removeRMSOffset(uint64_t sumSquared, int64_t sum, unsigned long numSamples);
double removePowerOffset(uint64_t power, int64_t sumV, int64_t sumI, unsigned long numSamples);
double x1, y1;  // phase shift coefficients
double applyPhaseShift(double phaseShift, double sampleRate, double A, double B);
double deg_rad(double a);

float Vrms, Irms;
long sumTimerCount;
float realPower,apparentPower,powerFactor;
float frequency;
volatile word timerCount=TIMERTOP;
volatile word pllUnlocked=PLLLOCKCOUNT;
word sumCycleCount;
volatile bool newsumCycle;
unsigned long nextTransmitTime;
bool rfmXmit = false;


static void showString (PGM_P s);

void setup()
{
  pinMode(LEDPIN, OUTPUT);
  digitalWrite(LEDPIN, HIGH);
  #ifdef EMONTX_V34
  //READ DIP SWITCH 1 POSITION 
  pinMode(DIP_SWITCH1, INPUT_PULLUP);
  //READ DIP SWITCH 2 POSITION 
  pinMode(DIP_SWITCH2, INPUT_PULLUP);
  #endif
  #ifdef SYNCPIN
  pinMode(SYNCPIN, OUTPUT);
  digitalWrite(SYNCPIN, LOW);
  #endif
  #ifdef SAMPPIN
  pinMode(SAMPPIN, OUTPUT);
  digitalWrite(SAMPPIN, LOW);
  #endif

  
  for (byte i=0; i<4; i++)
  {
      digitalWrite(LEDPIN, LOW); delay(200); digitalWrite(LEDPIN, HIGH); delay(200);
  }
    
 // start the SPI library:
  SPI.begin();
  SPI.setBitOrder(MSBFIRST);
  SPI.setDataMode(0);
  SPI.setClockDivider(SPI_CLOCK_DIV8);
    
  Serial.begin(9600);                            // Do NOT set greater than 9600
  digitalWrite(LEDPIN, LOW);   
  Serial.println(F("OpenEnergyMonitor.org"));
  #ifdef EMONTX_V2
   Serial.print(F("emonTx V2 "));
  #endif
  #ifdef EMONTX_V32
   Serial.print(F("emonTx V3.2 "));
  #endif
  #ifdef EMONTX_V34
   Serial.print(F("emonTx V3.4 "));
  #endif
  #ifdef EMONTX_SHIELD    
   Serial.print(F("emonTx Shield "));
  #endif    
   Serial.println(F(VERSION));



  calculateConstants();
  calculateTiming(SUPPLY_FREQUENCY);
 

  #ifdef DEBUGGING
    Serial.println(F("Phase shift coefficients:"));
    Serial.print(F("x1 = "));Serial.print(x1);Serial.print(F("  y1 = "));Serial.println(y1);
  #endif
  
  // change ADC prescaler to /64 = 250kHz clock
  // slightly out of spec of 200kHz but should be OK
  ADCSRA &= 0xf8;  // remove bits set by Arduino library
  ADCSRA |= 0x06; 

  //set timer 1 interrupt for required sumPeriod
  noInterrupts();
  TCCR1A = 0; // clear control registers
  TCCR1B = 0;
  TCNT1  = 0; // clear counter
  OCR1A = TIMERTOP; // set compare reg for timer sumPeriod
  bitSet(TCCR1B,WGM12); // CTC mode
  bitSet(TCCR1B,CS10); // no prescaling
  bitSet(TIMSK1,OCIE1A); // enable timer 1 compare interrupt
  bitSet(ADCSRA,ADIE); // enable ADC interrupt
  interrupts();

  nextTransmitTime=millis() + looptime;
  
  showHelptext();
}

void loop()
{
  getSettings();
  
  if(newsumCycle && !firstCycle)
    addsumCycle(); // a new mains sumCycle has been sampled
  firstCycle = false;
  
  if((millis()>=nextTransmitTime) && ((millis()-nextTransmitTime)<0x80000000L)) // check for overflow
  {
    #ifndef LEDISLOCK
      digitalWrite(LEDPIN,HIGH);
    #else
      digitalWrite(LEDPIN,LOW);
    #endif
    calculateVIPF();
    sendResults();
    nextTransmitTime+=looptime;
    #ifndef LEDISLOCK
      digitalWrite(LEDPIN,LOW);
    #else
      digitalWrite(LEDPIN,HIGH);
    #endif
  }
}

// timer 1 interrupt handler
ISR(TIMER1_COMPA_vect)
{
  #ifdef SAMPPIN
    digitalWrite(SAMPPIN,HIGH);
  #endif
  ADMUX = _BV(REFS0) | ctPin; // start ADC conversion for first current
  ADCSRA |= _BV(ADSC);
  #ifdef SAMPPIN
    digitalWrite(SAMPPIN,LOW);
  #endif
}

// ADC interrupt handler
ISR(ADC_vect)
{
  static int newV, lastV, sampleI;
  static int storedV[BUFFERSIZE];  // Array to store voltage samples
  int result;
  static int Vindex = 0;
  
  #ifdef SAMPPIN
    digitalWrite(SAMPPIN,HIGH);
  #endif
  result = ADCL;
  result |= ADCH<<8;
  // remove the nominal offset 
  result -=(ADC_COUNTS >> 1);
  // determine which conversion just completed
  
  if ((ADMUX & 0x0f) == ctPin)
  {
      ADMUX = _BV(REFS0) | VOLTSPIN; // start CT4 conversion
      ADCSRA |= _BV(ADSC);
      sampleI = result;
      sumIsq += (long)sampleI * sampleI;
      sumIavg += sampleI; 
  }
  
  else if ((ADMUX & 0x0f) == VOLTSPIN)
  {
      lastV=newV;
      newV = result;
      storedV[Vindex] = newV;        // store this voltage sample in circular buffer
      sumVsq += ((long)newV * newV);
      sumVavg += newV; 

      sumPowerA += (long)storedV[(Vindex+BUFFERSIZE-CTPhase)%BUFFERSIZE] * sampleI;  // Use stored & delayed voltage for power calculation
      sumPowerB += (long)storedV[(Vindex+BUFFERSIZE-CTPhase-1)%BUFFERSIZE] * sampleI;

      sumSamples++;
      updatePLL(newV,lastV);
      ++Vindex %= BUFFERSIZE;
  }
  #ifdef SAMPPIN
    digitalWrite(SAMPPIN,LOW);
  #endif

}

void updatePLL(int newV, int lastV)
{
  static byte samples=0;
  static int oldV;
  bool rising;

  rising=(newV>lastV); // synchronise to rising zero crossing
  
  samples++;
  if(samples>=NUMSAMPLES) // end of one 50Hz sumCycle
  {
    #ifdef SYNCPIN
      digitalWrite(SYNCPIN,HIGH);
    #endif
    samples=0;
    if(rising)
    {
      // if we're in the rising part of the 50Hz sumCycle adjust the final timer count
      // to move newV towards 0, only adjust if we're moving in the wrong direction
      if(((newV<0)&&(newV<=oldV))||((newV>0)&&(newV>=oldV))) timerCount-=newV;
      // limit range of PLL frequency
      timerCount=constrain(timerCount,PLLTIMERMIN,PLLTIMERMAX);
      OCR1A=timerCount;
      if(abs(newV)>PLLUNLOCKRANGE) pllUnlocked=PLLLOCKCOUNT; // we're unlocked
      else if(pllUnlocked) pllUnlocked--;
      #ifdef LEDISLOCK
        digitalWrite(LEDPIN,pllUnlocked?LOW:HIGH);
      #endif
    }
    else // in the falling part of the sumCycle, we shouldn't be here
    {
      OCR1A=PLLTIMERMAX; // shift out of this region fast
      pllUnlocked=PLLLOCKCOUNT; // and we can't be locked
    }
    
    oldV=newV;
    
    newsumCycle=true; // flag new sumCycle to outer loop
  }
  else if(samples==(NUMSAMPLES/2))
  {
    // negative zero crossing
    #ifdef SYNCPIN
      digitalWrite(SYNCPIN,LOW);
    #endif
  }
  #ifdef SAMPPIN
    digitalWrite(SAMPPIN,LOW);
  #endif
}

// add data for new 50Hz sumCycle to total for the period (called from loop() )
void addsumCycle()
{
  // save results for outer loop
  noInterrupts();
  sumPeriodVsq     += sumVsq;
  sumPeriodVavg    += sumVavg;
  sumPeriodIsq     += sumIsq;
  sumPeriodIavg    += sumIavg; 

  sumPeriodPowerA  += sumPowerA;
  sumPeriodPowerB  += sumPowerB;
  sumPeriodSamples += sumSamples;

  sumVsq	   = 0;
  sumVavg	   = 0;
  sumIsq	   = 0;
  sumIavg	   = 0; 
      
  sumPowerA	 = 0;
  sumPowerB	 = 0;
  sumSamples = 0;

  sumTimerCount+=(timerCount+1); // for average frequency calculation
  sumCycleCount++;
  newsumCycle=false;
  interrupts();
}


double removeRMSOffset(uint64_t sumSquared, int64_t sum, unsigned long numSamples)
{
    double x = ((double)sumSquared / numSamples) - ((double)sum * sum / numSamples / numSamples);
    return (x<0.0 ? 0.0 : sqrt(x));
}

double removePowerOffset(int64_t power, int64_t sumV, int64_t sumI, unsigned long numSamples)
{
    return (((double)power / numSamples) - ((double)sumV * sumI / numSamples / numSamples));
}

double deg_rad(double a)
{
    return (0.01745329*a);
}

double applyPhaseShift(double phaseShift, double sampleRate, double A, double B)
{
  double y = sin(deg_rad(phaseShift)) / sin(deg_rad(sampleRate));
  double x = cos(deg_rad(phaseShift)) - y * cos(deg_rad(sampleRate));
  return (A * x + B * y); 
}


// calculate voltage, current, power and frequency
void calculateVIPF()
{  
  if(sumPeriodSamples==0) return; // just in case
  
  frequency=((float)sumCycleCount*16000000)/(((float)sumTimerCount)*NUMSAMPLES);

  calculateTiming(frequency);

  // rms values - voltage & current
  
  // Vrms still contains the fine voltage offset. Correct this by subtracting the "Offset V^2" before the sq. root.
  Vrms = v_ratio * removeRMSOffset(sumPeriodVsq, sumPeriodVavg, sumPeriodSamples);
   
  // Similarly the 4 currents
  Irms = i_ratio * removeRMSOffset(sumPeriodIsq, sumPeriodIavg, sumPeriodSamples);

  // Power contains both voltage & current offsets. Correct this by subtracting the "Offset Power": Vavg * Iavg.
  // Apply timing/phase compensation to obtain real power.
  //  real power = Ical * Vcal * (powerA * PHASECAL - powerB * (PHASECAL - 1));
  // or more accurately:
  //  y = sin(phase_shift) / sin(sampleRate);
  //  x = cos(phase_shift) - y * cos(sampleRate);
  // realPower = Ical * Vcal * (powerA * x + powerB * y);
  // [sampleRate] is the angle between sample sets in radians - and is different for 
  //   50 Hz and 60 Hz systems
  // [phase_shift] will vary according to the time delay between the current and voltage samples
  //   as well as the difference in phase leads of the two transformers.
  // x & y have been calculated in setup() as they won't change.

  realPower = v_ratio * i_ratio * (x1 * removePowerOffset(sumPeriodPowerA, sumPeriodVavg, sumPeriodIavg, sumPeriodSamples) 
                                   + y1 * removePowerOffset(sumPeriodPowerB, sumPeriodVavg, sumPeriodIavg, sumPeriodSamples));
  apparentPower = Irms * Vrms; 
  if (apparentPower > 0.1)        // suppress "nan" values
    powerFactor = realPower / apparentPower;
  else
    powerFactor = 0.0;
  
   
  sumPeriodVsq     = 0;
  sumPeriodVavg    = 0;
  sumPeriodIsq     = 0;
  sumPeriodIavg    = 0;

  sumPeriodPowerA  = 0;
  sumPeriodPowerB  = 0;
  sumPeriodSamples = 0;

  sumCycleCount=0;
  sumTimerCount=0;

}

void calculateConstants(void)
{
  // Intermediate calculations

  v_ratio = vCal  * SUPPLY_VOLTS / 1024; 
  i_ratio = iCal  * SUPPLY_VOLTS / 1024;


}

void calculateTiming(float frequency)
{
  // Pre-calculate the constants for phase/timing correction
  phaseshift = (ADC_RATE * 3.6e-4 * frequency - iLead -iError); // in degrees
  y1 = sin(deg_rad(phaseshift)) / sin(deg_rad(SAMPLERATE));
  x1 = cos(deg_rad(phaseshift)) - y1 * cos(deg_rad(SAMPLERATE));
}

void sendResults()
{
  #ifdef TXPIN
    digitalWrite(TXPIN,HIGH);
    delay(10);
    digitalWrite(TXPIN,LOW);
  #endif

   
    Serial.print(F("Vrms: "));
    Serial.print(Vrms);
    Serial.print(F(" I: "));
    Serial.print(Irms);
   Serial.print(F(" f: "));
    Serial.print(frequency,3);

    Serial.print(F(" "));
    if(pllUnlocked) Serial.print(F(" PLL is unlocked "));
    else 
    {
      Serial.print(F(" PLL is locked "));
      Serial.print(F(" pf: "));
      Serial.print(powerFactor,4);
      Serial.print(F(" phase: "));
      Serial.print(iLead,3);
    }
    Serial.println();
  
}

